"""
Provides functionality for lazily deserializing classes field-by-field.

When deserializing an object, rather than deserializing all of its fields
immediately, some additional fields/methods are added:

- `_uniserde_field_definitions_`: A dictionary mapping field names to tuples of
  (field document name, field type).

- `_uniserde_remaining_fields_`: A dictionary mapping field document names to
  unparsed field values.

- `_uniserde_cache_`: The cache instance that was used to deserialize the
  object.

- `__getattr__`: A function that deserializes fields when they are first
  accessed, and then caches them in the instance.

Note that this approach allows for doing a lot of work up-front: The class,
along with all of its fields, and the `__getattr__` method only have to be
created once and can be reused for all future deserializations of that class.
"""

from __future__ import annotations

import inspect
import typing as t

from . import serde_cache, utils
from .errors import SerdeError
from .type_hint import TypeHint


def _lazy_getattr(self, name: str) -> t.Any:
    # Fetch the field definitions. This will fail if this particular instance
    # isn't lazy
    try:
        field_definitions: dict[str, tuple[str, TypeHint]] = vars(self)[
            "_uniserde_field_definitions_"
        ]
    except KeyError:
        raise AttributeError(name) from None

    # See if there is a field definition for this field. This will fail if the
    # field doesn't exist
    try:
        field_doc_name, field_type = field_definitions[name]
    except KeyError:
        raise AttributeError(name) from None

    # Get the field value
    try:
        field_raw_value = self._uniserde_remaining_fields_.pop(field_doc_name)
    except KeyError:
        raise SerdeError(f"Missing field {field_doc_name!r}") from None

    # Deserialize it
    field_handler = self._uniserde_cache_._get_handler(
        field_type,
    )
    parsed_value = field_handler(
        self._uniserde_cache_,
        field_raw_value,
        field_type,
    )

    # Cache it
    vars(self)[name] = parsed_value

    # Return it
    return parsed_value


def _get_attribute_map(
    as_type: TypeHint,
    serde: serde_cache.SerdeCache,
) -> dict[str, tuple[str, TypeHint]]:
    # Already cached?
    try:
        return serde._field_map_cache[as_type.origin]
    except KeyError:
        pass

    # Fetch the attributes
    result = {
        attribute_py_name: (
            serde._python_attribute_name_to_doc_name(attribute_py_name),
            attribute_type,
        )
        for (
            attribute_py_name,
            attribute_type,
        ) in utils.get_class_attributes_recursive(as_type.origin).items()
    }

    # Cache the result
    serde._field_map_cache[as_type.origin] = result

    # Return it
    return result


def can_create_lazy_instance(as_type: TypeHint) -> bool:
    """
    Verify that some conditions are met for creating a lazy instance of the
    given type.
    """

    # The class mustn't have a `__getattr__` method, since that would be
    # overwritten. If it is already overwritten that's fine though.
    try:
        return as_type.origin.__getattr__ is _lazy_getattr
    except AttributeError:
        return True


def create_lazy_instance(
    serialized: dict[str, t.Any],
    serdeserializer: serde_cache.SerdeCache,
    as_type: TypeHint,
) -> t.Any:
    assert isinstance(serialized, dict), serialized
    assert inspect.isclass(as_type.origin), as_type
    assert serdeserializer._lazy, serdeserializer
    assert can_create_lazy_instance(as_type), as_type

    # Instantiate the result, skipping the constructor
    result = object.__new__(as_type.origin)

    # Set additional, internal fields
    type(result).__getattr__ = _lazy_getattr
    result._uniserde_field_definitions_ = _get_attribute_map(as_type, serdeserializer)
    result._uniserde_cache_ = serdeserializer
    result._uniserde_remaining_fields_ = serialized

    # Done
    return result



from __future__ import annotations

import inspect
import types
import typing as t
import uuid
from datetime import datetime, timedelta

from .objectid_proxy import ObjectId


class TypeHint:
    """
    Represents a type hint with its origin and arguments.

    Unlike plain type hints these are easily comparable and support parsing
    themselves from string type hints.
    """

    def __init__(
        self,
        type_hint: t.Type | str,
        *,
        scope: dict[str, t.Any] | None = None,
    ) -> None:
        """
        Initialize the TypeHint instance.

        The origin and arguments are extracted from the given type hint. If any
        of the types are strings, they will be evaluated in the given scope.

        Some standardizations are performed on the type hint:

        - `types.UnionType` becomes a `Union` with the original arguments
        - `Union[T, None]` becomes `Optional[T]`

        ## Parameters

        `type_hint`: The type hint to be processed.

        ## Raises

        `TypeError`: If the type hint isn't supported by `uniserde` (e.g.
            `Callable`).

        `TypeError`: If one of the type hints was a string but no scope was
            provided.

        `TypeError`: If one of the type hints was a string and has raised an
            exception during evaluation.
        """
        # If the type hint is a string, evaluate it
        if isinstance(type_hint, str):
            if scope is None:
                raise TypeError(
                    "A scope must be provided to evaluate string type hints."
                )

            type_hint = TypeHint.evaluate_type_hint(type_hint, scope)

        # Get the origin
        self._origin: t.Type
        origin = t.get_origin(type_hint)
        args = t.get_args(type_hint)

        # Some built-in types like `str` or `int` are not recognized by
        # `get_origin`
        if origin is None:
            origin = type_hint

        # Perform standardization
        if origin is types.UnionType:
            origin = t.Union

        # Convert `Union[T, None]` to `Optional[T]`
        if origin is t.Union and len(args) == 2 and type(None) in args:
            origin = t.Optional
            args = tuple(arg for arg in args if arg is not type(None))

        # Catch unsupported type hints
        is_user_class = inspect.isclass(origin) and origin.__module__ != "builtins"

        if (
            origin
            not in (
                bool,
                int,
                float,
                bytes,
                str,
                tuple,
                list,
                set,
                dict,
                t.Optional,
                datetime,
                timedelta,
                ObjectId,
                uuid.UUID,
                t.Any,
                t.Literal,
            )
            and not is_user_class
        ):
            raise TypeError(f"`{origin}` is not supported by `uniserde`.")

        # Single argument type hints
        if origin in (list, set, t.Optional):
            if len(args) != 1:
                raise TypeError(f"Expected one argument for `{origin}`, got {args!r}.")

        # Dict (the only dual argument type hint)
        elif origin is dict:
            if len(args) != 2:
                raise TypeError(f"Expected two arguments for `dict`, got {args!r}.")

            if args[0] not in (str, int):
                raise TypeError(
                    f"`dict` keys must be of type `str` or `int`, not {args[0]!r}."
                )

        # Any number of arguments type hints
        elif origin in (tuple, t.Literal):
            pass

        # Zero argument type hints
        elif args:
            raise TypeError(f"Unexpected arguments for `{origin}`: {args!r}")

        # Store the origin
        self._origin = origin  # type: ignore

        # Get the args
        self._args: tuple[TypeHint, ...]
        self._literal_args: tuple[str, ...]

        if self._origin is t.Literal:
            for arg in args:
                if not isinstance(arg, str):
                    raise TypeError(
                        f"`uniserde` currently only supports string literals, not `{type_hint}`."
                    )

            self._args = tuple()
            self._literal_args = args

        else:
            parsed_args: list[TypeHint] = []

            for arg in args:
                parsed_args.append(TypeHint(arg, scope=scope))

            self._args = tuple(parsed_args)
            self._literal_args = tuple()

    @staticmethod
    def evaluate_type_hint(
        type_hint_str: str,
        scope: dict[str, t.Any],
    ) -> t.Type:
        """
        Evaluate a string representation of a type hint.

        ## Raises

        `TypeError`: If the string could not be evaluated.
        """
        try:
            return eval(type_hint_str, scope)
        except Exception as e:
            raise TypeError(f"Could not evaluate string type hint: {e}")

    @property
    def origin(self) -> t.Type:
        """
        Return the origin of the type hint.
        """
        return self._origin

    @property
    def args(self) -> tuple[TypeHint, ...]:
        """
        Return the arguments of the type hint.
        """
        return self._args

    @property
    def literal_args(self) -> tuple[str, ...]:
        """
        Return the literal arguments of the type hint.
        """
        return self._literal_args

    def as_python(self) -> t.Type:
        """
        Returns the type hint as a Python type. For example a type hint with
        origin `list` and arguments `(TypeHint(int))` will return `List[int]`.
        """
        # If not a generic, just return the origin
        if not self._args and not self._literal_args:
            assert not self._args, (self._origin, self._args)
            assert not self._literal_args, (self._origin, self._literal_args)

            return self._origin

        # Special case: Literals
        if self._origin is t.Literal:
            args = self.literal_args
        else:
            args = [arg.as_python() for arg in self._args]

        # Pass in the arguments, or not?
        if len(args) == 0:
            return self._origin  # type: ignore

        if len(args) == 1:
            return self._origin[args[0]]  # type: ignore

        return self._origin[tuple(args)]  # type: ignore

    def __hash__(self) -> int:
        """
        Return the hash of the TypeHint instance.

        ## Returns

        An integer hash value.
        """
        return hash((self._origin, self._args))

    def __eq__(self, other: t.Any) -> bool:
        """
        Compare the TypeHint instance with another instance for equality.

        ## Parameters

        `other`: The other instance to compare with.

        ## Returns

        `True` if both instances are equal, `False` otherwise.
        """
        if not isinstance(other, TypeHint):
            return False

        return self._origin == other._origin and self._args == other._args

    def pretty_string(self) -> str:
        # Special case: Literals
        if self._origin is t.Literal:
            return f'Literal[{", ".join(self.literal_args)}]'

        # Type without arguments
        if not self._args:
            return self._origin.__name__

        # Type with arguments
        args = [arg.pretty_string() for arg in self._args]
        return f"{self._origin.__name__}[{', '.join(args)}]"

    def __repr__(self) -> str:
        return f"TypeHint({self.pretty_string()})"



"""
This module provides the `ObjectId` class. If the `bson` module is available, it
is imported from there. Otherwise it falls back to a class with an identical -
albeit incomplete - interface.

`uniserde` expects the `bson` package provided by `pymongo`, NOT the one called
`bson` on PyPI. However, both seem to work fine.
"""

from __future__ import annotations

import binascii
import os
import struct
import time
import typing as t

import typing_extensions as te

__all__ = [
    "ObjectId",
]


class _InvalidId(ValueError):
    """Raised when an invalid ObjectId is encountered."""


class _ObjectId:
    """
    This is an incomplete clone of `bson.ObjectId`. Use this as a stand-in
    if the `bson` library isn't available. While this class aims to be
    similar to the original, it does not provide the same stringent
    uniqueness guarantees.
    """

    def __init__(
        self,
        oid: str | t.Literal[ObjectId] | bytes | None = None,  # type: ignore
    ) -> None:
        if oid is None:
            self._id_blob = struct.pack(">I", int(time.time())) + os.urandom(8)

        elif isinstance(oid, bytes) and len(oid) == 12:
            self._id_blob = oid

        elif isinstance(oid, str) and len(oid) == 24:
            try:
                self._id_blob = bytes.fromhex(oid)
            except (TypeError, ValueError):
                raise ValueError(f"Invalid hex in ObjectId: {oid}")

        else:
            raise TypeError(
                f"ObjectIds must be 12 bytes or 24-character hex strings. Got {oid!r}"
            )

    @property
    def binary(self) -> bytes:
        """12-byte binary representation of this ObjectId."""
        return self._id_blob

    def __str__(self) -> str:
        return binascii.hexlify(self._id_blob).decode()

    def __repr__(self) -> str:
        return f"ObjectId('{str(self)}')"

    def __eq__(self, other: t.Any) -> bool:
        if isinstance(other, ObjectId):
            return self._id_blob == other.binary

        return NotImplemented

    def __hash__(self) -> int:
        return hash(self._id_blob)


try:
    import bson  # type: ignore

    ObjectId: te.TypeAlias = bson.ObjectId  # type: ignore
    errors = bson.errors  # type: ignore
except ImportError:
    ObjectId: te.TypeAlias = _ObjectId  # type: ignore

    class errors:
        InvalidId = _InvalidId  # type: ignore



"""
Given a uniserde data model, generates the Rust equivalent for it. This assumes
Rust uses the `serde` library for serialization / deserialization. It doesn't
produce perfect code, but the results are typically a good starting point for
writing your own, final version.

The code will be written to `uniserde_models.rs` in the current working
directory.
"""

import enum
import inspect
import typing as t
import uuid
from dataclasses import dataclass
from datetime import datetime, timedelta
from pathlib import Path

import uniserde
import uniserde.case_convert
import uniserde.type_hint
import uniserde.utils
from uniserde.type_hint import TypeHint

PYTHON_TYPES_TO_RUST_TYPES = {
    bool: "bool",
    int: "i32",
    float: "f64",
    str: "String",
    bytes: "Vec<u8>",
    datetime: "DateTime<Utc>",
    timedelta: "Duration",
    uuid.UUID: "Uuid",
}


@dataclass
class SubModel:
    id: str
    timestamp: datetime
    reference_to_other_instance: str
    amount: int
    description: str
    events: list[str]


@dataclass
class Model:
    name: str
    subs: list[SubModel]
    sets: set[str]
    timestamp: datetime
    deleted: bool
    duration: float


def get_all_model_classes(py_type: t.Type) -> t.Iterable[type]:
    """
    Given a model class, returns an iterable over all model classes that must be
    generated.
    """
    as_type = TypeHint(py_type)

    # These types already exist in Rust and need no further processing
    if as_type.origin in {bool, int, float, str, bytes, datetime}:
        return

    # Classes are themselves models, and may contain more
    if inspect.isclass(as_type.origin) and as_type.origin not in {
        tuple,
        list,
        set,
        dict,
    }:
        yield as_type.origin

        for _, attr_as_type in uniserde.utils.get_class_attributes_recursive(
            as_type.origin
        ).items():
            yield from get_all_model_classes(attr_as_type.origin)

    # Recur into the args
    for arg_type in as_type.args:
        yield from get_all_model_classes(arg_type.origin)


def write_rust_struct(
    model: TypeHint,
    out: t.TextIO,
    *,
    public: bool,
) -> None:
    """
    Given a model class, writes the Rust equivalent as a struct.
    """
    assert isinstance(model, TypeHint), model
    assert inspect.isclass(model.origin), model

    out.write("#[derive(Debug, Deserialize)]\n")

    if public:
        out.write("pub ")

    out.write(f"struct {model.origin.__name__} {{\n")

    for ii, (field_name, attr_as_type) in enumerate(
        uniserde.utils.get_class_attributes_recursive(model.origin).items()
    ):
        if ii != 0:
            out.write("\n")

        # Special cases
        if attr_as_type.origin is datetime:
            out.write(f'    #[serde(deserialize_with = "deserialize_iso8601")]\n')
            out.write(f"    {field_name}: DateTime<Utc>,\n")
            continue

        if attr_as_type.origin is timedelta:
            out.write(
                f'    #[serde(deserialize_with = "deserialize_duration_from_seconds")]\n'
            )
            out.write(f"    {field_name}: Duration,\n")
            continue

        # General case
        field_rust_type = convert_type_to_rust(attr_as_type)
        out.write(f"    {field_name}: {field_rust_type},\n")

    out.write("}\n")


def write_rust_enum(
    model: TypeHint,
    out: t.TextIO,
    *,
    public: bool,
) -> None:
    """
    Given an enum class, writes the Rust equivalent as an enum.
    """
    assert issubclass(model.origin, enum.Enum), model

    out.write("#[derive(Debug, Deserialize)]\n")

    if public:
        out.write("pub ")

    out.write(f"enum {model.origin.__name__} {{\n")

    for member in model.origin:
        out.write(f'    #[serde(rename = "{member.name}")]\n')
        out.write(
            f"    {uniserde.case_convert.all_upper_to_upper_camel_case(member.name)},\n\n"
        )

    out.write("}\n")


def write_rust_model(
    model: TypeHint,
    out: t.TextIO,
    *,
    public: bool = True,
) -> None:
    if issubclass(model.origin, enum.Enum):
        write_rust_enum(model, out, public=public)
    else:
        write_rust_struct(model, out, public=public)


def convert_type_to_rust(as_type: TypeHint) -> str:
    """
    Given a Python type, returns the equivalent Rust type.
    """
    # Simple lookups
    try:
        return PYTHON_TYPES_TO_RUST_TYPES[as_type.origin]
    except KeyError:
        pass

    # More complex types
    if as_type.origin is list:
        return f"Vec<{convert_type_to_rust(as_type.args[0])}>"

    if as_type.origin is tuple:
        return f"({', '.join(convert_type_to_rust(arg) for arg in as_type.args)})"

    if as_type.origin is set:
        return f"HashSet<{convert_type_to_rust(as_type.args[0])}>"

    if as_type.origin is dict:
        return f"HashMap<{convert_type_to_rust(as_type.args[0])}, {convert_type_to_rust(as_type.args[1])}>"

    if as_type.origin is t.Optional:
        return f"Option<{convert_type_to_rust(as_type.args[0])}>"

    if inspect.isclass(as_type.origin):
        return as_type.origin.__name__

    raise ValueError(f"Unsupported type: {as_type}")


def write_boilerplate(out: t.TextIO) -> None:
    out.write(
        """
use serde::Deserialize;
use chrono::{DateTime, Duration, Utc};


fn deserialize_iso8601<'de, D>(deserializer: D) -> Result<DateTime<Utc>, D::Error>
where
    D: serde::Deserializer<'de>,
{
    let s: &str = Deserialize::deserialize(deserializer)?;
    DateTime::parse_from_rfc3339(s)
        .map(|dt| dt.with_timezone(&Utc))
        .map_err(serde::de::Error::custom)
}


fn deserialize_duration_from_seconds<'de, D>(deserializer: D) -> Result<Duration, D::Error>
where
    D: serde::Deserializer<'de>,
{
    let seconds: f64 = Deserialize::deserialize(deserializer)?;
    // Convert seconds (float) to chrono::Duration
    Duration::from_std(std::time::Duration::from_secs_f64(seconds))
        .map_err(serde::de::Error::custom)
}
            """.strip()
    )


def main() -> None:
    # Find all models that need porting
    all_model_classes = set(get_all_model_classes(Model))

    with Path("uniserde_models.rs").open("w") as out:
        # Common header code
        write_boilerplate(out)
        out.write("\n\n")

        # Write the Model Code
        for model_cls in all_model_classes:
            write_rust_model(TypeHint(model_cls), out)
            out.write("\n\n")


if __name__ == "__main__":
    main()



"""
Generates code for serializing / deserializing the model class to/from JSON.
It doesn't produce perfect code, as the functions expect to get access to some
values that only exist at runtime. However, the results are typically a good
starting point for writing your own, final version.

The code will be written to `uniserde_json.py` in the current working directory.
"""

from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

import uniserde
from uniserde.type_hint import TypeHint


@dataclass
class SubModel:
    id: str
    timestamp: datetime
    reference_to_other_instance: str
    amount: int
    description: str
    events: list[str]


@dataclass
class Model:
    name: str
    subs: list[SubModel]
    sets: set[str]
    timestamp: datetime
    deleted: bool
    duration: float


def postprocess_code(
    code: str,
    method_name: str,
) -> str:
    # Cut off the header
    pos = code.find("):")
    assert pos != -1, "Cannot find the end of the function signature"

    code = code[pos + 2 :]

    # Replace silly names
    return code.replace("dateutil_parser", "dateutil.parser")


def main() -> None:
    # Prepare the code-generators for both serialization and deserialization
    #
    # Serialization
    serde = uniserde.JsonSerde()

    handler_builder = serde._serialization_cache._get_handler_builder(TypeHint(Model))
    ser_gen = serde._serialization_cache._write_code_from_handler_builder(
        handler_builder,
        TypeHint(Model),
    )

    # Deserialization
    handler_builder = serde._deserialization_cache._get_handler_builder(TypeHint(Model))
    deser_gen = serde._deserialization_cache._write_code_from_handler_builder(
        handler_builder,
        TypeHint(Model),
    )

    with Path("uniserde_json.py").open("w") as f:
        # Header text
        f.write(
            """
# This file was autogenerated by `uniserde`'s `write_model_code.py` script. It
# contains code for serializing / deserializing the model class to/from JSON.

from __future__ import annotations

from datetime import datetime, timedelta, timezone
import base64
import binascii
import uuid
import typing as t

import dateutil.parser


class SerdeError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)

    @property
    def message(self) -> str:
        return self.args[0]
""".strip()
        )
        f.write("\n\n\n")

        f.write("# Serialization\n")
        f.write("def as_json(self) -> dict[str, t.Any]:\n")
        f.write("    value_in = self\n")
        f.write(
            postprocess_code(
                ser_gen.resulting_code(),
                "as_json",
            )
        )
        f.write("\n\n\n")

        f.write("# Deserialization\n")
        f.write("def from_json(document: dict[str, t.Any]) -> Model:\n")
        f.write("    value_in = document\n")
        f.write(
            postprocess_code(
                deser_gen.resulting_code(),
                "from_json",
            )
        )
        f.write("\n")


if __name__ == "__main__":
    main()



import uniserde
import test.models as models
import uniserde.lazy_wrapper as lazy_wrapper
import uniserde.json_deserialize
import uniserde.bson_deserialize

value_fresh = models.TestClass.create_variant_1()
value_serialized = uniserde.as_json(value_fresh)

value_lazy: models.TestClass = lazy_wrapper.create_lazy_instance(
    value_serialized,
    uniserde.json_deserialize.JsonDeserializationCache(
        custom_handlers={},
        lazy=False,
    ),
    models.TestClass,
)

print(value_lazy.val_str)
print(value_lazy.val_datetime)
print(value_lazy.val_doesnt_exist)



from __future__ import annotations


import uniserde
import uniserde.compat
from datetime import datetime, timezone
import typing as t
import dataclasses
from bson import ObjectId
import tests.models as models


serde = uniserde.BsonSerde(lazy=False)

value_fresh = models.ChildClass.create_child_variant_1()

value_bson = serde.as_bson(value_fresh)
assert value_bson["type"] == "ChildClass"

value_round_trip = serde.from_bson(
    models.ParentClass,
    value_bson,
)

assert isinstance(value_round_trip, models.ChildClass)
assert value_fresh == value_round_trip

raise SystemExit()


from dataclasses import dataclass
import uniserde
import uniserde.codegen
import uuid
import tests.models
import enum


original = tests.models.ClassWithStaticmethodOverrides.create()


print(f"Original: {original}")

serialized = uniserde.as_bson(original)

print(f"Serialized: {serialized}")

deserialized = uniserde.from_bson(
    type(original),
    serialized,
)

print(f"Deserialized: {deserialized}")



from __future__ import annotations

import re

_SPLIT_CAMEL_CASE_PATTERN = re.compile(
    ".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)"
)


def _is_lower(c: str) -> bool:
    """
    Does the obvious.

    The built-in `str.islower` method is not used because it mysteriously
    returns `False` for some strings, such as `_`.
    """
    return c == c.lower()


def _is_upper(c: str) -> bool:
    """
    Does the obvious.

    The built-in `str.islower` method is not used because it mysteriously
    returns `False` for some strings, such as `_`.
    """
    return c == c.upper()


def all_lower_to_camel_case(name: str) -> str:
    """
    Converts a string from all_lower_case to camelCase.
    """
    assert _is_lower(name), name

    if not name:
        return ""

    parts = name.split("_")
    assert parts, (name, parts)

    return parts[0] + "".join(p.capitalize() for p in parts[1:])


def all_lower_to_camel_case_with_id_exception(name: str) -> str:
    """
    Convert all lower case names to camel case, except for the string "id",
    which becomes "_id".
    """
    if name == "id":
        return "_id"

    return all_lower_to_camel_case(name)


def identity(name: str) -> str:
    """
    Keeps the name the same.
    """
    return name


def all_upper_to_camel_case(name: str) -> str:
    """
    Converts a string from ALL_UPPER_CASE to camelCase.
    """
    assert _is_upper(name), name

    if not name:
        return ""

    parts = name.split("_")
    assert parts, (name, parts)

    return parts[0].lower() + "".join(p.capitalize() for p in parts[1:])


def all_upper_to_upper_camel_case(name: str) -> str:
    """
    Converts a string from ALL_UPPER_CASE to UpperCamelCase.
    """
    assert _is_upper(name), name

    if not name:
        return ""

    parts = name.split("_")
    assert parts, (name, parts)

    return "".join(p.capitalize() for p in parts)


def camel_case_to_all_upper(name: str) -> str:
    """
    Converts a string from camelCase to ALL_UPPER_CASE.
    """
    matches = _SPLIT_CAMEL_CASE_PATTERN.finditer(name)
    groups = [m.group(0).upper() for m in matches]
    return "_".join(groups)


def upper_camel_case_to_camel_case(name: str) -> str:
    """
    Converts a string from UpperCamelCase to camelCase.
    """
    if not name:
        return ""

    return name[0].lower() + name[1:]


def all_lower_to_kebab_case(name: str) -> str:
    """
    Converts a string from all_lower_case to kebab-case.
    """
    assert _is_lower(name), name
    return name.replace("_", "-")



from __future__ import annotations

import pytest

import tests.models as models
import uniserde.lazy_wrapper


def test_missing_attribute_raises_attribute_error() -> None:
    serde = uniserde.JsonSerde(lazy=True)

    # Create a non-lazy instance
    value_eager = models.TestClass.create_variant_1()

    # Create a lazy instance of the same class
    value_lazy = serde.from_json(
        models.TestClass,
        serde.as_json(value_eager),
    )

    assert type(value_eager).__getattr__ is uniserde.lazy_wrapper._lazy_getattr  # type: ignore

    # This has modified the class' `__getattr__` method. Now test that accessing
    # an invalid attribute in fact raises an `AttributeError`, rather than just
    # failing, e.g. because the class is missing some fields expected to be
    # contained in lazy instances.
    with pytest.raises(AttributeError):
        value_lazy.this_is_an_invalid_attribute_name  # type: ignore



from __future__ import annotations

import base64
import binascii
import enum
import inspect
import pathlib
import typing as t
import uuid
from datetime import datetime, timedelta
from pathlib import Path

import dateutil.parser

from . import (
    codegen,
    lazy_wrapper,
    objectid_proxy,
    serde_cache,
    utils,
)
from .objectid_proxy import ObjectId
from .type_hint import TypeHint
from .typedefs import Jsonable


def _build_passthrough_handler(
    serde: JsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    simple_type_hint: TypeHint,
) -> str:
    """
    Builds a handler that simply checks the type of the value and raises an
    exception if it is not the expected type.

    The value itself is returned as-is.
    """
    gen.write(
        f"if not isinstance({value}, {simple_type_hint.origin.__name__}):",
        f"    raise SerdeError('Expected {simple_type_hint.origin.__name__}, got {{}}'.format({value}))",
    )

    return value


def _build_handler_int_from_int(
    serde: JsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> str:
    gen.write(
        f"if not isinstance({value}, int) and not (isinstance({value}, float) and {value}.is_integer()):",
        f"    raise SerdeError('Expected int, got {{}}'.format({value}))",
    )
    return f"int({value})"


def _build_handler_float_from_float(
    serde: JsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> str:
    gen.write(
        f"if not isinstance({value}, (int, float)):",
        f"    raise SerdeError('Expected float, got {{}}'.format({value!r}))",
    )
    return f"float({value})"


def _build_handler_bytes_from_str(
    serde: JsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> str:
    result_var = gen.get_new_variable()

    gen.expose_value("base64", base64)
    gen.expose_value("binascii", binascii)

    gen.write(
        f"if not isinstance({value}, str):",
        f"    raise SerdeError('Expected bytes encoded as base64, got {{}}'.format({value}))",
        f"",
        f"try:",
        f"    {result_var} = base64.b64decode({value}.encode('utf-8'))",
        f"except binascii.Error as e:",
        f"    raise SerdeError('Encountered invalid base64 string: {{}}'.format(e)) from None",
    )

    return result_var


def _build_handler_datetime_from_str(
    serde: JsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> str:
    result_var = gen.get_new_variable()

    gen.expose_value("dateutil_parser", dateutil.parser)

    gen.write(
        f"if not isinstance({value}, str):",
        f"    raise SerdeError('Expected date/time string, got {{}}'.format({value}))",
        f"",
        f"try:",
        f"  {result_var} = dateutil_parser.isoparse({value})",
        f"except ValueError:",
        f"  raise SerdeError('Invalid date/time string: {{}}'.format({value})) from None",
        f"",
        f"if {result_var}.tzinfo is None:",
        f"    raise SerdeError('The date/time value `{{}}` is missing a timezone.'.format({value}))",
    )

    return result_var


def _build_handler_float_to_timedelta(
    serde: JsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> str:
    result_var = gen.get_new_variable()

    gen.expose_value("timedelta", timedelta)

    gen.write(
        f"if not isinstance({value}, (int, float)):",
        f"    raise SerdeError('Expected duration in seconds, got {{}}'.format({value}))",
        f"",
        f"{result_var} = timedelta(seconds={value})",
    )

    return result_var


def _build_handler_list_to_tuple(
    serde: JsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> str:
    gen.write(
        f"if not isinstance({value}, list):",
        f"    raise SerdeError('Expected list, got {{}}'.format({value}))",
        f"",
        f"if len({value}) != {len(as_type.args)}:",
        f"    raise SerdeError('Expected list of length {len(as_type.args)}, got {{}}'.format(len({value})))",
        f"",
    )

    # Convert the individual values
    subresults: list[str] = []

    for ii, sub_type in enumerate(as_type.args):
        subresult = serde._write_single_handler(
            gen,
            f"{value}[{ii}]",
            sub_type,
        )
        subresults.append(subresult)

    # Return the result
    result_var = gen.get_new_variable()

    gen.write("", f"{result_var} = tuple([{', '.join(subresults)}])")

    return result_var


def _build_handler_list_to_list(
    serde: JsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> str:
    result_var = gen.get_new_variable()
    count_var = gen.get_new_variable()

    gen.write(
        f"if not isinstance({value}, list):",
        f"    raise SerdeError('Expected list, got {{}}'.format({value}))",
        f"",
        f"{result_var} = []",
        f"",
        f"for {count_var} in {value}:",
    )

    gen.indentation_level += 1

    subresult = serde._write_single_handler(
        gen,
        count_var,
        as_type.args[0],
    )

    gen.write(f"{result_var}.append({subresult})")

    gen.indentation_level -= 1
    return result_var


def _build_handler_list_to_set(
    serde: JsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> str:
    result_var = gen.get_new_variable()
    count_var = gen.get_new_variable()

    gen.write(
        f"if not isinstance({value}, list):",
        f"    raise SerdeError('Expected list, got {{}}'.format({value}))",
        f"",
        f"{result_var} = set()",
        f"",
        f"for {count_var} in {value}:",
    )

    gen.indentation_level += 1

    subresult = serde._write_single_handler(
        gen,
        count_var,
        as_type.args[0],
    )

    gen.write(f"{result_var}.add({subresult})")

    gen.indentation_level -= 1
    return result_var


def _build_handler_dict_to_dict(
    serde: JsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> str:
    result_var = gen.get_new_variable()
    key_var = gen.get_new_variable()
    value_var = gen.get_new_variable()

    key_as_type, value_as_type = as_type.args

    gen.write(
        f"if not isinstance({value}, dict):",
        f"    raise SerdeError('Expected dict, got {{}}'.format({value}))",
        f"",
        f"{result_var} = {{}}",
        f"",
        f"for {key_var}, {value_var} in {value}.items():",
    )

    gen.indentation_level += 1

    # Deserialize the key
    if key_as_type.origin is str:
        key_result = f"{key_var}"
    elif key_as_type.origin is int:
        key_result = gen.get_new_variable()
        gen.write(
            f"try:",
            f"    {key_result} = int({key_var})",
            f"except ValueError:",
            f'    raise SerdeError("Invalid key in object: Expected integer, got {{}}".format({key_var})) from None',
        )
    else:
        raise AssertionError(
            f"{key_as_type.pretty_string()} is not supported for dictionary keys"
        )

    # Deserialize the value
    value_result = serde._write_single_handler(
        gen,
        value_var,
        value_as_type,
    )

    gen.write(f"{result_var}[{key_result}] = {value_result}")

    gen.indentation_level -= 1
    return result_var


def _build_handler_object_id_from_str(
    serde: JsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> str:
    result_var = gen.get_new_variable()

    gen.expose_value("bson", objectid_proxy)

    gen.write(
        f"if not isinstance({value}, str):",
        f"    raise SerdeError('Expected str, got {{}}'.format({value}))",
        f"",
        f"try:",
        f"    {result_var} = bson.ObjectId({value})",
        f"except bson.errors.InvalidId as e:",
        f"    raise SerdeError('Invalid ObjectId: {{}}'.format(e)) from None",
    )

    return result_var


def _build_handler_optional(
    serde: JsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> t.Any:
    # Don't get too clever here. Yes, it would be nice to reuse the same result
    # variable as the subresult, but that would not only require a needless
    # negation in the `if`, but also lead to problems if a subresult doesn't
    # actually return a variable, but say `int(variable)`.
    result_var = gen.get_new_variable()

    gen.write(
        f"if {value} is None:",
        f"    {result_var} = None",
        f"else:",
    )

    gen.indentation_level += 1
    subresult_var = serde._write_single_handler(
        gen,
        value,
        as_type.args[0],
    )
    gen.write(f"{result_var} = {subresult_var}")
    gen.indentation_level -= 1

    return result_var


def _build_handler_any_to_any(
    serde: JsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> t.Any:
    return value


def _build_handler_literal_from_str(
    serde: JsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> t.Any:
    gen.write(
        f"if not isinstance({value}, str):",
        f"    raise SerdeError('Expected str, got {{}}'.format({value}))",
        f"",
        f"if {value} not in {as_type.literal_args!r}:",
        f'    raise SerdeError("Expected one of {{!r}}, got {{}}".format({repr(as_type.args), value}))',
    )

    return value


def _build_handler_path_from_str(
    serde: JsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> str:
    result_var = gen.get_new_variable()

    gen.expose_value("pathlib", pathlib)

    gen.write(
        f"if not isinstance({value}, str):",
        f"    raise SerdeError('Expected str, got {{}}'.format({value}))",
        f"",
        f"{result_var} = pathlib.Path({value})",
    )

    return result_var


def _build_handler_uuid_from_str(
    serde: JsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> str:
    result_var = gen.get_new_variable()

    gen.expose_value("uuid", uuid)

    gen.write(
        f"if not isinstance({value}, str):",
        f"    raise SerdeError('Expected str, got {{}}'.format({value}))",
        f"",
        f"try:",
        f"    {result_var} = uuid.UUID({value})",
        f"except ValueError as e:",
        f"    raise SerdeError('Invalid UUID: {{}}'.format(e)) from None",
    )

    return result_var


JSON_HANDLER_BUILDERS: dict[t.Type, serde_cache.HandlerBuilder] = {
    bool: _build_passthrough_handler,
    int: _build_handler_int_from_int,
    float: _build_handler_float_from_float,
    str: _build_passthrough_handler,
    bytes: _build_handler_bytes_from_str,
    datetime: _build_handler_datetime_from_str,
    timedelta: _build_handler_float_to_timedelta,
    tuple: _build_handler_list_to_tuple,
    list: _build_handler_list_to_list,
    set: _build_handler_list_to_set,
    dict: _build_handler_dict_to_dict,
    t.Optional: _build_handler_optional,
    t.Any: _build_handler_any_to_any,
    ObjectId: _build_handler_object_id_from_str,
    t.Literal: _build_handler_literal_from_str,
    Path: _build_handler_path_from_str,
    type(Path()): _build_handler_path_from_str,
    uuid.UUID: _build_handler_uuid_from_str,
}  # type: ignore


class JsonDeserializationCache(serde_cache.SerdeCache[Jsonable, t.Any]):
    """
    Configuration & cache for deserializing JSON into Python objects.
    """

    def __init__(
        self,
        *,
        context: t.Any,
        custom_handlers: dict[t.Type, serde_cache.InternalHandler],
        lazy: bool,
        python_attribute_name_to_doc_name: t.Callable[[str], str],
        python_class_name_to_doc_name: t.Callable[[str], str],
        python_enum_name_to_doc_name: t.Callable[[str], str],
    ) -> None:
        super().__init__(
            context=context,
            eager_class_handler_builders=JSON_HANDLER_BUILDERS,
            lazy_class_handler_builders={},
            override_method_name="_uniserde_from_json_",
            user_provided_handlers=custom_handlers,
            lazy=lazy,
            python_attribute_name_to_doc_name=python_attribute_name_to_doc_name,
            python_class_name_to_doc_name=python_class_name_to_doc_name,
            python_enum_name_to_doc_name=python_enum_name_to_doc_name,
        )

    def _build_attribute_by_attribute_class_handler(
        self,
        gen: codegen.Codegen,
        input_variable_name: str,
        as_type: TypeHint,
    ) -> str:
        assert inspect.isclass(as_type.origin), as_type

        # Make sure the input is a dictionary. By handling it here the code
        # isn't repeated in every subclass.
        gen.write(
            f"if not isinstance({input_variable_name}, dict):",
            f"    raise SerdeError('Expected class instance stored as object, got {{}}'.format({input_variable_name}))",
            f"",
        )

        # If this class is not serialized `@as_child`, create regular
        # deserialization logic.
        result_var = gen.get_new_variable()
        serialize_as_root = utils.root_of_serialize_as_child(as_type.origin)

        if serialize_as_root is None:
            self._build_attribute_by_attribute_class_handler_without_children(
                gen,
                input_variable_name,
                result_var,
                as_type,
            )
            return result_var

        # Otherwise precompute a list of possible classes. Then delegate to a
        # deserializer for every possible child class.
        doc_key_to_child_class = {
            self._python_class_name_to_doc_name(sub_cls.__name__): sub_cls
            for sub_cls in utils.all_subclasses(serialize_as_root, True)
        }

        # Which class to deserialize is stored in the `type` field
        type_var = gen.get_new_variable()

        gen.write(
            f"try:",
            f'    {type_var} = {input_variable_name}.pop("type")',
            f"except KeyError:",
            f'    raise SerdeError("Missing `type` field in {as_type.origin.__name__} instance") from None',
            f"",
            f"match {type_var}:",
        )
        gen.indentation_level += 1

        for sub_cls_type_key, sub_cls in doc_key_to_child_class.items():
            gen.write(
                f"case {sub_cls_type_key!r}:",
            )
            gen.indentation_level += 1

            self._build_attribute_by_attribute_class_handler_without_children(
                gen,
                input_variable_name,
                result_var,
                TypeHint(sub_cls),
            )

            gen.indentation_level -= 1

        # Error case
        gen.write(
            f"case _:",
            f'    raise SerdeError("Invalid `type` value in {as_type.origin.__name__} instance: {{}}".format({type_var}))',
        )

        gen.indentation_level -= 1

        # Phew!
        return result_var

    def _build_attribute_by_attribute_class_handler_without_children(
        self,
        gen: codegen.Codegen,
        input_variable_name: str,
        result_var: str,
        as_type: TypeHint,
    ) -> None:
        dict_var = gen.get_new_variable()

        # Go lazy?
        if self._lazy and lazy_wrapper.can_create_lazy_instance(as_type):
            lazy_wrapper_var = gen.get_new_variable()
            self_var = gen.get_new_variable()
            cls_var = gen.get_new_variable()

            gen.expose_value(lazy_wrapper_var, lazy_wrapper)
            gen.expose_value(self_var, self)
            gen.expose_value(cls_var, as_type)

            gen.write(
                f"{result_var} = {lazy_wrapper_var}.create_lazy_instance({input_variable_name}, {self_var}, {cls_var})",
            )
            return

        # Go eager!
        cls_var = gen.get_new_variable()
        gen.expose_value(cls_var, as_type.origin)

        gen.write(
            f"{result_var} = object.__new__({cls_var})",
            f"{dict_var} = vars({result_var})",
            f"",
        )

        # Deserialize all fields
        for field_py_name, field_as_type in utils.get_class_attributes_recursive(
            as_type.origin
        ).items():
            field_doc_name = self._python_attribute_name_to_doc_name(field_py_name)

            gen.write(f"# {field_py_name}")
            field_var = gen.get_new_variable()
            gen.write(
                f"try:",
                f"    {field_var} = {input_variable_name}.pop({field_doc_name!r})",
                f"except KeyError:",
                f"    raise SerdeError('Missing field {{}}'.format({field_doc_name!r})) from None",
                f"",
            )

            subresult = self._write_single_handler(
                gen,
                field_var,
                field_as_type,
            )
            gen.write(f"{dict_var}[{field_py_name!r}] = {subresult}")

        # Make sure no superfluous fields are present
        gen.write(
            f"",
            f"if {input_variable_name}:",
            f"    raise SerdeError('Object contains superfluous fields: {{}}'.format({input_variable_name}.keys()))",
        )

    def _build_flag_enum_handler(
        self,
        gen: codegen.Codegen,
        input_variable_name: str,
        as_type: TypeHint,
    ) -> str:
        assert issubclass(as_type.origin, enum.Flag), as_type

        result_var = gen.get_new_variable()
        cls_var = gen.get_new_variable()
        options_var = gen.get_new_variable()
        count_var = gen.get_new_variable()

        gen.expose_value(cls_var, as_type.origin)

        # Prepare a serialized version of all options. Do this right in the
        # code, so the enum options can already be instantiated.
        gen.write(f"{options_var} = {{")

        for option in as_type.origin:
            # How can opt_py_type be None here? According to VSCode it can be
            assert option.name is not None, option
            option_py_name = option.name
            option_doc_name = self._python_enum_name_to_doc_name(option_py_name)

            gen.write(f"    {option_doc_name!r}: {cls_var}.{option_py_name},")

        # Look up all received options and add them to the result
        gen.write(
            f"}}",
            f"",
            f"{result_var} = {cls_var}(0)",
            f"",
            f"for {count_var} in {input_variable_name}:",
            f"    if not isinstance({count_var}, str):",
            f"        raise SerdeError('Expected enumeration value as string, got `{{}}`'.format({count_var}))",
            f"",
            f"    try:",
            f"        {result_var} |= {options_var}[{count_var}]",
            f"    except KeyError:",
            f"        raise SerdeError('Invalid enumeration value `{{}}`'.format({count_var})) from None",
        )

        return result_var

    def _build_enum_handler(
        self,
        gen: codegen.Codegen,
        input_variable_name: str,
        as_type: TypeHint,
    ) -> str:
        assert issubclass(as_type.origin, enum.Enum), as_type

        result_var = gen.get_new_variable()
        cls_var = gen.get_new_variable()
        options_var = gen.get_new_variable()

        gen.expose_value(cls_var, as_type.origin)

        # Prepare a serialized version of all options. Do this right in the
        # code, so the enum options can already be instantiated.
        gen.write(f"{options_var} = {{")

        for option in as_type.origin:
            option_py_name = option.name
            option_doc_name = self._python_enum_name_to_doc_name(option_py_name)

            gen.write(f"    {option_doc_name!r}: {cls_var}.{option_py_name},")

        # Look up the value
        gen.write(
            f"}}",
            f"",
            f"if not isinstance({input_variable_name}, str):",
            f"    raise SerdeError('Expected enumeration value as string, got `{{}}`'.format({input_variable_name}))",
            f"",
            f"try:",
            f"    {result_var} = {options_var}[{input_variable_name}]",
            f"except KeyError:",
            f"    raise SerdeError('Invalid enumeration value `{{}}`'.format({input_variable_name})) from None",
        )

        return result_var



from __future__ import annotations

import typing as t
import uuid
from datetime import datetime

import typing_extensions as te

from .objectid_proxy import ObjectId

__all__ = [
    "Jsonable",
    "Bsonable",
    "JsonDoc",
    "BsonDoc",
]


Jsonable: te.TypeAlias = t.Union[
    None,
    bool,
    int,
    float,
    str,
    dict[str, "Jsonable"],
    list["Jsonable"],
    tuple["Jsonable", ...],
]

Bsonable: te.TypeAlias = t.Union[
    None,
    bool,
    int,
    float,
    str,
    dict[str, "Bsonable"],
    list["Bsonable"],
    tuple["Bsonable", ...],
    bytes,
    ObjectId,
    datetime,
    uuid.UUID,
]

JsonDoc = dict[str, Jsonable]
BsonDoc = dict[str, Bsonable]



import typing as t
import uuid
from datetime import datetime, timedelta

import pytest

from uniserde import ObjectId
from uniserde.type_hint import TypeHint

UNPARSED_VALID_TYPE_HINTS = (
    (bool, "bool", bool, ()),
    (int, "int", int, ()),
    (float, "float", float, ()),
    (str, "str", str, ()),
    (tuple[int], "tuple[int]", tuple, (int,)),
    (list[str], "list[str]", list, (str,)),
    (set[float], "set[float]", set, (float,)),
    (dict[str, int], "dict[str, int]", dict, (str, int)),
    (t.Optional[str], "t.Optional[str]", t.Optional, (str,)),
    (t.Union[str, None], "t.Union[str, None]", t.Optional, (str,)),
    (t.Union[None, str], "t.Union[None, str]", t.Optional, (str,)),
    (datetime, "datetime", datetime, ()),
    (timedelta, "timedelta", timedelta, ()),
    (ObjectId, "ObjectId", ObjectId, ()),
    (uuid.UUID, "uuid.UUID", uuid.UUID, ()),
)

UNPARSED_INVALID_TYPE_HINTS = (
    None,
    type(None),
    t.Optional,
    list,
    set,
    dict,
    dict[int],  # type: ignore  (intentional error for testing)
    t.Union,
    t.Union[int, str],
)


@pytest.mark.parametrize(
    "raw_type_hint, _, expected_origin, expected_args",
    UNPARSED_VALID_TYPE_HINTS,
)
def test_instantiate_parsed_type_hint(
    raw_type_hint: t.Type,
    _: str,
    expected_origin: t.Type,
    expected_args: tuple[t.Type, ...],
) -> None:
    """
    Instantiates a `TypeHint` object and checks the origin and args.
    """
    # Construct the `TypeHint` object
    type_hint_object = TypeHint(raw_type_hint)

    # Check the origin
    assert type_hint_object.origin is expected_origin

    # Check the args
    assert len(type_hint_object.args) == len(expected_args)

    for arg_should, arg_is in zip(expected_args, type_hint_object.args):
        assert arg_should is arg_is.origin


@pytest.mark.parametrize(
    "_, raw_type_hint, expected_origin, expected_args",
    UNPARSED_VALID_TYPE_HINTS,
)
def test_instantiate_string_type_hint(
    _: t.Type,
    raw_type_hint: str,
    expected_origin: t.Type,
    expected_args: tuple[t.Type, ...],
) -> None:
    """
    Instantiates a `TypeHint` object from a string and checks the origin and
    args.
    """
    # Construct the `TypeHint` object. Note that the hint is converted to a
    # string first.
    type_hint_object = TypeHint(
        raw_type_hint,
        scope=globals(),
    )

    # Check the origin
    assert type_hint_object.origin is expected_origin

    # Check the args
    assert len(type_hint_object.args) == len(expected_args)

    for arg_should, arg_is in zip(expected_args, type_hint_object.args):
        assert arg_should is arg_is.origin


@pytest.mark.parametrize(
    "raw_type_hint",
    UNPARSED_INVALID_TYPE_HINTS,
)
def test_instantiate_invalid_type_hint(
    raw_type_hint: t.Type,
) -> None:
    """
    Instantiates a `TypeHint` object from an invalid type hint and checks that
    a `TypeError` is raised.
    """
    with pytest.raises(TypeError):
        TypeHint(raw_type_hint)



from __future__ import annotations

import typing as t
import warnings

import pytest

import tests.models as models
import uniserde

# This throws a warning since 3.11, which shows up rather confusingly in pytest.
# Suppress it. (The warning is caused by `mongo_schema`, not uniserde.)
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import mongo_schema

try:
    import bson  # type: ignore
except ImportError:
    bson = None


def get_bson_serde_without_uuid_checks() -> uniserde.BsonSerde:
    """
    `mongo_schema` doesn't consider UUIDs to be valid `binData`. While perfectly
    reasonable, in `pymongo`, UUIDs do result in `binData` when stored in the
    database.

    This function builds a MongoDB schema but disables UUID verification to keep
    `mongo_schema` happy.
    """
    serde = uniserde.BsonSerde()
    serde._mongodb_schema_converter._enable_uuid_verification = False
    return serde


@pytest.mark.skipif(bson is None, reason="`pymongo` is not installed")
@pytest.mark.parametrize(
    "value, py_type",
    [
        (models.TestClass.create_variant_1(), models.TestClass),
        (models.TestClass.create_variant_2(), models.TestClass),
        (models.ParentClass.create_parent_variant_1(), models.ParentClass),
        (models.ChildClass.create_child_variant_1(), models.ChildClass),
        (models.ChildClass.create_child_variant_1(), models.ParentClass),
        (models.ClassWithId.create(), models.ClassWithId),
        (models.ClassWithKwOnly.create(), models.ClassWithKwOnly),
    ],
)
def test_value_matches_schema(value: t.Any, py_type: t.Type) -> None:
    serde = get_bson_serde_without_uuid_checks()

    schema = serde.as_mongodb_schema(py_type)

    bson_value = serde.as_bson(value)

    mongo_schema.validate(bson_value, schema)


def test_overridden_as_mongodb_schema_staticmethod() -> None:
    serde = get_bson_serde_without_uuid_checks()
    value_schema = serde.as_mongodb_schema(models.ClassWithStaticmethodOverrides)

    assert value_schema == {
        "value": "overridden value",
        "format": "mongodb schema",
    }


def test_overridden_as_mongodb_schema_classmethod() -> None:
    serde = get_bson_serde_without_uuid_checks()
    value_schema = serde.as_mongodb_schema(models.ClassWithClassmethodOverrides)

    assert value_schema == {
        "value": "overridden value",
        "format": "mongodb schema",
    }



"""
Compatibility module for loading/writing old-style uniserde JSON & BSON.

This module is semi-public. Feel free to use it, but it may change or be dropped
without a deprecation period or semver version bump.
"""

import typing as t
import warnings

import uniserde
from uniserde import Bsonable, BsonDoc, Jsonable, JsonDoc

T = t.TypeVar("T")


# Functions to help with updating old-school documents
def bulk_update_json(
    old_document: t.Iterable[JsonDoc],
    as_type: t.Type,
    *,
    new_serde: uniserde.JsonSerde | None,
) -> t.Iterable[JsonDoc]:
    """
    Reads a sequence of old-school JSON documents and returns them as a sequence
    of new-style JSON documents (in the same order).

    If `new_serde` is not provided, a default instance of `JsonSerde` will be
    used.
    """
    # Prepare the serde instances
    old_serde = uniserde.JsonSerde.new_camel_case()

    if new_serde is None:
        new_serde = uniserde.JsonSerde()

    # Walk the documents
    for doc in old_document:
        # Deserialize the old document
        obj = old_serde.from_json(as_type, doc)

        # Serialize the new document
        new_doc = new_serde.as_json(obj)
        assert isinstance(new_doc, dict)

        yield new_doc


def bulk_update_bson(
    old_document: t.Iterable[BsonDoc],
    as_type: t.Type,
    *,
    new_serde: uniserde.BsonSerde | None,
) -> t.Iterable[BsonDoc]:
    """
    Reads a sequence of old-school BSON documents and returns them as a sequence
    of new-style BSON documents (in the same order).

    If `new_serde` is not provided, a default instance of `BsonSerde` will be
    used.
    """
    # Prepare the serde instances
    old_serde = uniserde.BsonSerde.new_camel_case()

    if new_serde is None:
        new_serde = uniserde.BsonSerde()

    # Walk the documents
    for doc in old_document:
        # Deserialize the old document
        obj = old_serde.from_bson(as_type, doc)

        # Serialize the new document
        new_doc = new_serde.as_bson(obj)
        assert isinstance(new_doc, dict)

        yield new_doc


async def update_mongodb_collection(
    *,
    new_serde: uniserde.BsonSerde,
    motor_collection,
    as_type: t.Type,
) -> None:
    """
    Updates all documents in a MongoDB collection from the old-school format to
    the new-school format.

    If `new_serde` is not provided, a default instance of `BsonSerde` will be
    used.
    """

    # Prepare the serde instances
    old_serde = uniserde.BsonSerde.new_camel_case()

    # Walk the documents
    async for doc in motor_collection.find():
        # Deserialize the old document
        obj = old_serde.from_bson(as_type, doc)

        # Serialize the new document
        new_doc = new_serde.as_bson(obj)
        assert isinstance(new_doc, dict)
        assert "_id" in new_doc, f'No "_id" field in new document: {new_doc}'

        # Update the document
        await motor_collection.replace_one({"_id": doc["_id"]}, new_doc)


# Keep the old API alive
def as_json(
    value: t.Any,
    *,
    as_type: t.Optional[t.Type] = None,
    custom_serializers: dict[t.Type, t.Callable[[t.Any], t.Any]] = {},
) -> JsonDoc:
    warnings.warn("This method is deprecated.", DeprecationWarning)
    assert not custom_serializers, "Custom serializers are not supported here"
    serde = uniserde.JsonSerde.new_camel_case()
    return serde.as_json(value, as_type=as_type)


def from_json(
    document: t.Any,
    as_type: t.Type,
    *,
    custom_deserializers: dict[t.Type, t.Callable[[Jsonable], t.Any]] = {},
    lazy: bool = False,
) -> t.Any:
    warnings.warn("This method is deprecated.", DeprecationWarning)
    assert not custom_deserializers, "Custom deserializers are not supported here"
    serde = uniserde.JsonSerde.new_camel_case()
    return serde.from_json(as_type, document)


def as_bson(
    value: t.Any,
    *,
    as_type: t.Optional[t.Type] = None,
    custom_serializers: dict[t.Type, t.Callable[[t.Any], t.Any]] = {},
) -> BsonDoc:
    warnings.warn("This method is deprecated.", DeprecationWarning)
    assert not custom_serializers, "Custom serializers are not supported here"
    serde = uniserde.BsonSerde.new_camel_case()
    return serde.as_bson(value, as_type=as_type)


def from_bson(
    document: t.Any,
    as_type: t.Type,
    *,
    custom_deserializers: dict[t.Type, t.Callable[[Bsonable], t.Any]] = {},
    lazy: bool = False,
) -> t.Any:
    warnings.warn("This method is deprecated.", DeprecationWarning)
    assert not custom_deserializers, "Custom deserializers are not supported here"
    serde = uniserde.BsonSerde.new_camel_case()
    return serde.from_bson(as_type, document)


class Serde:
    def __init_subclass__(cls) -> None:
        warnings.warn(
            "This class is deprecated - there is no direct replacement in the new API.",
            DeprecationWarning,
        )

    def as_bson(
        self,
        *,
        as_type: t.Optional[t.Type] = None,
        custom_serializers: dict[t.Type, t.Callable[[t.Any], t.Any]] = {},
    ) -> BsonDoc:
        return as_bson(self, as_type=as_type, custom_serializers=custom_serializers)

    def as_json(
        self,
        *,
        as_type: t.Optional[t.Type] = None,
        custom_serializers: dict[t.Type, t.Callable[[t.Any], t.Any]] = {},
    ) -> JsonDoc:
        return as_json(self, as_type=as_type, custom_serializers=custom_serializers)

    @classmethod
    def from_bson(
        cls: t.Type[T],
        document: BsonDoc,
        *,
        custom_deserializers: dict[t.Type, t.Callable[[t.Any], t.Any]] = {},
        lazy: bool = False,
    ) -> T:
        return from_bson(
            document, as_type=cls, custom_deserializers=custom_deserializers, lazy=lazy
        )

    @classmethod
    def from_json(
        cls: t.Type[T],
        document: JsonDoc,
        *,
        custom_deserializers: dict[t.Type, t.Callable[[t.Any], t.Any]] = {},
        lazy: bool = False,
    ) -> T:
        warnings.warn(
            "This method is deprecated. Use the module-level `from_json` function instead.",
            DeprecationWarning,
        )
        return from_json(
            document,
            as_type=cls,
            custom_deserializers=custom_deserializers,
            lazy=lazy,
        )

    @classmethod
    def as_mongodb_schema(
        cls,
        *,
        custom_handlers: dict[t.Type, t.Callable[[t.Any], t.Any]] = {},
    ) -> t.Any:
        raise NotImplementedError(
            "This method is not supported in the compatibility API"
        )



from __future__ import annotations
import typing_extensions as te
import typing as t

from pathlib import Path
import uniserde

import uniserde.case_convert


class Config:
    """
    Easy way to load and save configuration files in TOML format.

    The configuration file is loaded once, and then cached. Any modified values
    are remembered and written back in one go later on. Writing changes to the
    file is done atomically and preserves comments and formatting in the file.
    """

    # Where to load the config from and save it back to
    uniserde_config_path: Path | None

    # Keeps track of any keys which were changed and need to be written back
    # to the file
    _uniserde_dirty_keys: set[str]

    def __init_subclass__(cls) -> None:
        # This class depends on `tomlkit`. Make sure it's installed
        try:
            import tomlkit  # type: ignore
        except ImportError:
            raise RuntimeError(
                "`uniserde.Config` requires the `tomlkit` package to be installed."
            ) from None

        # Make sure all fields have default values, i.e. the class can be
        # instantiated without any arguments.
        try:
            cls()
        except TypeError:
            raise TypeError(
                f"`{cls.__name__}` must be instantiable without any arguments. Make sure all fields have default values."
            ) from None

    def __init__(self) -> None:
        self.uniserde_config_path = None
        self._uniserde_dirty_keys = set()

    @classmethod
    def uniserde_load(cls, path: Path) -> te.Self:
        """
        Creates a new instance of this class, populating it with the contents of
        the TOML file at the given path.

        Any missing or invalid fields are replaced with the fields' default
        values.
        """
        import tomlkit  # type: ignore

        # Try to load the file from disk
        try:
            with path.open() as f:
                raw_values = tomlkit.load(f).unwrap()

        except Exception:
            raw_values = {}

        # Instantiate the result with default values
        self = cls()

        # Deserialize each field individually, imputing defaults as needed
        for py_field_name, field_type in uniserde.get_global_class_attributes(
            cls
        ).items():
            doc_field_name = uniserde.case_convert.all_lower_to_kebab_case(
                py_field_name
            )

            try:
                raw_field = raw_values[doc_field_name]
                field_value = uniserde.from_json(raw_field, field_type)
            except (KeyError, uniserde.SerdeError):
                pass
            else:
                setattr(self, py_field_name, field_value)

        # Setting all of the fields will make the result think that they are
        # dirty. Clear that flag.
        self._uniserde_dirty_keys.clear()

        # Done!
        return self

    def uniserde_dump(self) -> None:
        """
        Write the cached config values back to the `self.config_path` file.
        """
        assert self.uniserde_config_path is not None
        import tomlkit  # type: ignore
        import tomlkit.exceptions  # type: ignore

        # Make sure the parent directory exists
        self.uniserde_config_path.parent.mkdir(parents=True, exist_ok=True)

        # Fetch an up-to-date copy of the file contents, with all formatting
        # intact
        full_dump = uniserde.as_json(self)

        try:
            with self.uniserde_config_path.open() as f:
                new_toml_dict = tomlkit.load(f)

        # If it can't be read, preserve all known values
        except (OSError, tomlkit.exceptions.TOMLKitError):
            new_toml_dict = tomlkit.TOMLDocument()

            for doc_field_name, value in full_dump.items():
                new_toml_dict[doc_field_name] = value

        # Otherwise add just the dirty keys
        else:
            for py_field_name in self._uniserde_dirty_keys:
                doc_field_name = uniserde.case_convert.all_lower_to_kebab_case(
                    py_field_name
                )
                new_toml_dict[doc_field_name] = full_dump[doc_field_name]

        # Dump the new TOML
        with self.uniserde_config_path.open("w") as f:
            tomlkit.dump(new_toml_dict, f)

        # All values have been saved and are now clean
        self._uniserde_dirty_keys.clear()

    def __setattr__(self, name: str, value: t.Any) -> None:
        # Store the new value
        self.__dict__[name] = value

        # Remember that this key was changed
        self._uniserde_dirty_keys.add(name)

    def __enter__(self) -> "Config":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        # Make sure to write any changes back to the toml file
        self.uniserde_dump()



from __future__ import annotations

import typing as t
import uuid
from datetime import datetime

from . import (
    codegen,
    json_serialize,
    objectid_proxy,
    serde_cache,
)
from .objectid_proxy import ObjectId
from .type_hint import TypeHint
from .typedefs import Bsonable


def _build_handler_datetime_to_datetime(
    serde: BsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> str:
    gen.expose_value("datetime", datetime)

    gen.write(
        f"assert isinstance({value}, datetime), {value}",
        f'assert {value}.tzinfo is not None, "Encountered datetime without a timezone. Please always set timezones, or expect hard to find bugs."',
    )
    return value


def _build_handler_objectid_to_objectid(
    serde: BsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> str:
    gen.expose_value("bson", objectid_proxy)

    gen.write(
        f"assert isinstance({value}, bson.ObjectId), {value}",
    )
    return value


def _build_handler_uuid_to_uuid(
    serde: BsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: TypeHint,
) -> str:
    gen.expose_value("uuid", uuid)

    gen.write(
        f"assert isinstance({value}, uuid.UUID), {value}",
    )
    return value


BSON_HANDLER_BUILDERS = {
    **json_serialize.JSON_HANDLER_BUILDERS,
    bytes: json_serialize._build_passthrough_handler,
    datetime: _build_handler_datetime_to_datetime,
    ObjectId: _build_handler_objectid_to_objectid,
    uuid.UUID: _build_handler_uuid_to_uuid,
}


class BsonSerializationCache(serde_cache.SerdeCache[Bsonable, t.Any]):
    """
    Configuration & cache for serializing BSON into Python objects.
    """

    def __init__(
        self,
        *,
        context: t.Any,
        custom_handlers: dict[t.Type, serde_cache.InternalHandler],
        python_attribute_name_to_doc_name: t.Callable[[str], str],
        python_class_name_to_doc_name: t.Callable[[str], str],
        python_enum_name_to_doc_name: t.Callable[[str], str],
    ) -> None:
        super().__init__(
            context=context,
            eager_class_handler_builders=BSON_HANDLER_BUILDERS,
            lazy_class_handler_builders={},
            override_method_name="_uniserde_as_bson_",
            user_provided_handlers=custom_handlers,
            lazy=False,
            python_attribute_name_to_doc_name=python_attribute_name_to_doc_name,
            python_class_name_to_doc_name=python_class_name_to_doc_name,
            python_enum_name_to_doc_name=python_enum_name_to_doc_name,
        )

    _build_attribute_by_attribute_class_handler = json_serialize.JsonSerializationCache._build_attribute_by_attribute_class_handler  # type: ignore

    _build_attribute_by_attribute_class_handler_without_children = json_serialize.JsonSerializationCache._build_attribute_by_attribute_class_handler_without_children  # type: ignore

    _build_flag_enum_handler = (
        json_serialize.JsonSerializationCache._build_flag_enum_handler
    )

    _build_enum_handler = json_serialize.JsonSerializationCache._build_enum_handler



import copy
import timeit
import uniserde.lazy_wrapper
from dataclasses import dataclass
from datetime import datetime, timezone
import typing as t

import uniserde


@dataclass
class SubModel:
    id: str
    timestamp: datetime
    reference_to_other_instance: str
    amount: int
    description: str
    events: list[str]


@dataclass
class Model:
    name: str
    subs: list[SubModel]
    sets: set[str]
    timestamp: datetime
    deleted: bool
    duration: float


RAW = {
    "name": "test",
    "subs": [
        {
            "id": "1",
            "timestamp": "2020-01-01T00:00:00Z",
            "referenceToOtherInstance": "2",
            "amount": 100,
            "description": "test",
            "events": ["a", "b", "c"],
        },
        {
            "id": "2",
            "timestamp": "2020-01-01T00:00:00Z",
            "referenceToOtherInstance": "1",
            "amount": 200,
            "description": "test",
            "events": ["a", "b", "c"],
        },
    ],
    "sets": ["a", "b", "c"],
    "timestamp": "2020-01-01T00:00:00Z",
    "deleted": False,
    "duration": 1.0,
}


@dataclass
@uniserde.as_child
class ParentClass(uniserde.Serde):
    parent_int: int
    parent_float: float

    @classmethod
    def create_parent_variant_1(cls) -> "ParentClass":
        return cls(
            parent_int=1,
            parent_float=1.0,
        )


@dataclass
class ChildClass(ParentClass):
    child_float: float
    child_str: str

    @classmethod
    def create_child_variant_1(cls) -> "ChildClass":
        return cls(
            parent_int=1,
            parent_float=1.0,
            child_float=1.0,
            child_str="this is a string",
        )


def main() -> None:
    as_json = ChildClass.create_child_variant_1().as_json()
    uniserde.from_json(ChildClass, as_json)

    number = 10000

    try:
        serde = uniserde.JsonSerde(
            lazy=True,
        )
    except AttributeError:

        def deserialize_lazy() -> None:
            uniserde.from_json(Model, RAW)
    else:

        def deserialize_lazy() -> None:
            uniserde.from_json(Model, RAW, ctx=serde)

    print(
        "deepcopy",
        timeit.timeit(
            lambda: copy.deepcopy(RAW),
            number=number,
        ),
    )

    print(
        "uniserde eager",
        timeit.timeit(
            lambda: uniserde.from_json(copy.deepcopy(RAW), Model),
            number=number,
        ),
    )

    print(
        "uniserde lazy",
        timeit.timeit(
            deserialize_lazy,
            number=number,
        ),
    )


if __name__ == "__main__":
    main()



from __future__ import annotations

import enum
import typing as t
import uuid
from dataclasses import KW_ONLY, dataclass
from datetime import datetime, timedelta, timezone
from pathlib import Path

import uniserde
from uniserde import JsonDoc, ObjectId


class RegularEnum(enum.Enum):
    ONE = 1
    TWO = 2
    THREE = 3


class FlagEnum(enum.Flag):
    ONE = 1
    TWO = 2
    FOUR = 4


@dataclass
class SimpleClass:
    foo: int
    bar: str


@dataclass
class TestClass:
    id: int

    val_bool: bool
    val_int: int
    val_float: float
    val_bytes: bytes
    val_str: str
    val_datetime: datetime
    val_timedelta: timedelta
    val_tuple: tuple[int, str]
    val_list: list[int]
    val_set: set[int]
    val_dict: dict[str, int]
    val_optional: t.Optional[int]
    val_old_union_optional_1: t.Union[int, None]
    val_old_union_optional_2: t.Union[None, int]
    val_new_union_optional_1: int | None
    val_new_union_optional_2: None | int
    val_any: t.Any
    val_object_id: ObjectId
    val_literal: t.Literal["one", "two", "three"]
    val_enum: RegularEnum
    val_flag: FlagEnum
    val_path: Path
    val_uuid: uuid.UUID
    val_class: SimpleClass

    @classmethod
    def create_variant_1(cls) -> TestClass:
        return cls(
            id=1,
            val_bool=True,
            val_int=1,
            val_float=1.0,
            val_bytes=b"these are bytes",
            val_str="this is a string",
            val_datetime=datetime(2020, 1, 2, tzinfo=timezone.utc),
            val_timedelta=timedelta(days=1, seconds=2, microseconds=3),
            val_tuple=(1, "one"),
            val_list=[1, 2, 3],
            val_set={1, 2, 3},
            val_dict={"one": 1, "two": 2},
            val_optional=1,
            val_old_union_optional_1=1,
            val_old_union_optional_2=1,
            val_new_union_optional_1=1,
            val_new_union_optional_2=1,
            val_any="this is an ANY value",
            val_object_id=ObjectId("62bd611fa847c71f1b68fffb"),
            val_literal="one",
            val_enum=RegularEnum.ONE,
            val_flag=FlagEnum.ONE | FlagEnum.TWO,
            val_path=Path.home() / "one",
            val_uuid=uuid.UUID("754a5dbf-e7f3-4cc3-b2d7-9382e586cfd3"),
            val_class=SimpleClass(foo=1, bar="one"),
        )

    @classmethod
    def create_variant_2(cls) -> TestClass:
        return cls(
            id=2,
            val_bool=False,
            val_int=2,
            val_float=2.0,
            val_bytes=b"these are different bytes",
            val_str="this is another string",
            val_datetime=datetime(2024, 5, 6, tzinfo=timezone.utc),
            val_timedelta=timedelta(days=10, seconds=20, microseconds=30),
            val_tuple=(2, "two"),
            val_list=[4, 5, 6],
            val_set={4, 5, 6},
            val_dict={"three": 3, "four": 4},
            val_optional=None,
            val_old_union_optional_1=None,
            val_old_union_optional_2=None,
            val_new_union_optional_1=None,
            val_new_union_optional_2=None,
            val_any="this is another ANY value",
            val_object_id=ObjectId("62bd6122a847c71f1b68fffc"),
            val_literal="two",
            val_enum=RegularEnum.TWO,
            val_flag=FlagEnum.ONE | FlagEnum.TWO | FlagEnum.FOUR,
            val_path=Path.home() / "two",
            val_uuid=uuid.UUID("0eadbc7e-3418-45a5-b276-53e7d91bb79d"),
            val_class=SimpleClass(foo=2, bar="two"),
        )


@dataclass
@uniserde.as_child
class ParentClass:
    parent_int: int
    parent_float: float

    @classmethod
    def create_parent_variant_1(cls) -> ParentClass:
        return cls(
            parent_int=1,
            parent_float=1.0,
        )

    def serialized_should(self) -> JsonDoc:
        return {
            "type": "ParentClass",
            "parent_int": self.parent_int,
            "parent_float": self.parent_float,
        }


@dataclass
class ChildClass(ParentClass):
    child_float: float
    child_str: str

    @classmethod
    def create_child_variant_1(cls) -> ChildClass:
        return cls(
            parent_int=1,
            parent_float=1.0,
            child_float=1.0,
            child_str="this is a string",
        )

    def serialized_should(self) -> JsonDoc:
        return {
            "type": "ChildClass",
            "parent_int": self.parent_int,
            "parent_float": self.parent_float,
            "child_float": self.child_float,
            "child_str": self.child_str,
        }


@dataclass
class ClassWithId:
    id: int
    foo: int

    @classmethod
    def create(cls) -> ClassWithId:
        return cls(1, 2)


@dataclass
class ClassWithKwOnly:
    foo: int

    _: KW_ONLY

    bar: int

    @classmethod
    def create(cls) -> ClassWithKwOnly:
        return cls(1, bar=2)


@dataclass
class ClassWithStaticmethodOverrides:
    """
    Class which has uniserde's special methods overridden. This allows to check
    that they are called rather than the default.

    All methods are overridden as @staticmethod.
    """

    value: str
    format: str

    @classmethod
    def create(cls) -> ClassWithStaticmethodOverrides:
        return cls("stored value", "python")

    def _uniserde_as_json_(
        self,
        serde: uniserde.JsonSerde,
        as_type: t.Type,
    ) -> uniserde.JsonDoc:
        assert isinstance(serde, uniserde.JsonSerde)
        assert as_type is ClassWithStaticmethodOverrides, as_type

        return {"value": "overridden during serialization", "format": "json"}

    def _uniserde_as_bson_(
        self,
        serde: uniserde.BsonSerde,
        as_type: t.Type,
    ) -> uniserde.BsonDoc:
        assert isinstance(serde, uniserde.BsonSerde)
        assert as_type is ClassWithStaticmethodOverrides

        return {"value": "overridden during serialization", "format": "bson"}

    @staticmethod
    def _uniserde_from_json_(
        document: dict[str, t.Any],
        serde: uniserde.JsonSerde,
        as_type: t.Type,
    ) -> ClassWithStaticmethodOverrides:
        assert isinstance(document, dict)
        assert isinstance(serde, uniserde.JsonSerde)
        assert as_type is ClassWithStaticmethodOverrides

        return ClassWithStaticmethodOverrides(
            "overridden during deserialization", "json"
        )

    @staticmethod
    def _uniserde_from_bson_(
        document: dict[str, t.Any],
        serde: uniserde.BsonSerde,
        as_type: t.Type,
    ) -> ClassWithStaticmethodOverrides:
        assert isinstance(document, dict)
        assert isinstance(serde, uniserde.BsonSerde)
        assert as_type is ClassWithStaticmethodOverrides

        return ClassWithStaticmethodOverrides(
            "overridden during deserialization", "bson"
        )

    @staticmethod
    def _uniserde_as_mongodb_schema_(
        serde: uniserde.BsonSerde,
        as_type: t.Type,
    ) -> t.Any:
        assert isinstance(serde, uniserde.BsonSerde)
        assert as_type is ClassWithStaticmethodOverrides

        return {"value": "overridden value", "format": "mongodb schema"}


@dataclass
class ClassWithClassmethodOverrides:
    """
    Same as the class above, but with the methods overridden as @classmethod.
    """

    value: str
    format: str

    @classmethod
    def create(cls) -> ClassWithClassmethodOverrides:
        return cls("stored value", "python")

    @classmethod
    def _uniserde_from_json_(
        cls,
        document: dict[str, t.Any],
        serde: uniserde.JsonSerde,
        as_type: t.Type,
    ) -> ClassWithClassmethodOverrides:
        assert isinstance(document, dict)
        assert isinstance(serde, uniserde.JsonSerde)
        assert as_type is ClassWithClassmethodOverrides

        return ClassWithClassmethodOverrides(
            "overridden during deserialization", "json"
        )

    @classmethod
    def _uniserde_from_bson_(
        cls,
        document: dict[str, t.Any],
        serde: uniserde.BsonSerde,
        as_type: t.Type,
    ) -> ClassWithClassmethodOverrides:
        assert isinstance(document, dict)
        assert isinstance(serde, uniserde.BsonSerde)
        assert as_type is ClassWithClassmethodOverrides

        return ClassWithClassmethodOverrides(
            "overridden during deserialization", "bson"
        )

    @classmethod
    def _uniserde_as_mongodb_schema_(
        cls,
        serde: uniserde.BsonSerde,
        as_type: t.Type,
    ) -> t.Any:
        assert isinstance(serde, uniserde.BsonSerde)
        assert as_type is ClassWithClassmethodOverrides

        return {"value": "overridden value", "format": "mongodb schema"}
